#include <windows.h>
#ifdef __BORLANDC__
#include <wincon.h>
#endif

// int getch();
// int getch2(int *shift, int *ctrl, int *vk);
// void BreakOff();

// These functions use the win32 api--no interrupts.

// getch() does not allow the keypress to show on the screen.
// It returns the ascii value of the key if is an ascii key.
// Non-ascii keys require using getch() twice.  The first return is zero,
// the second is the scan code as if from interrupt 16h function 10h.

// getch2() does not allow the keypress to show on the screen, either.
// It returns the ascii value of ascii keys and zero for other keys.
// It does not need to be called twice--its arguments contain the rest of
// the information.  "shift" and "ctrl" are flags for upper case and
// the ctrl-key state, "vk" ls the win32 VK_ scan code.

// If BreakOff() has been called, ctrl-c will be returned as ascii 3
// and will not interrupt execution.  Ctrl-break will still halt as usual.
// If BreakOff() is not called, ctrl-c is fatal.

//--------------implementation---------------

#define EXP __declspec(dllexport)

BOOL WINAPI EXP LibMain(HINSTANCE h, DWORD msg, LPVOID pdum)
  // a dll's "main" function
  {
  if(msg==DLL_PROCESS_ATTACH) return TRUE;
  if(msg==DLL_PROCESS_DETACH) return TRUE;
  return FALSE;
  }

void Break_Auxilliary()
  // helper for the ctrl-c handler
  // Create a keyboard event with ascii "3" waiting.
  // (Windows doesn't consider ctrl-c to be a keyboard event.)
  {
  HANDLE h;
  INPUT_RECORD inrec;
  KEY_EVENT_RECORD r;
  DWORD s;
  int nr, k, shp, cpl, done;
  r.bKeyDown=1;
  r.uChar.AsciiChar=3;
  inrec.EventType=KEY_EVENT;
  inrec.Event.KeyEvent=r;
  h=GetStdHandle(STD_INPUT_HANDLE);
  WriteConsoleInput(h,&inrec,1,&nr);
  }

BOOL __stdcall Break_Handler(DWORD event)
  // ctrl-c handler
  // Handle ctrl-c but pass ctrl-break, etc.
  {
  if(event==CTRL_C_EVENT) { Break_Auxilliary(); return 1; }
  return 0;
  }

void EXP BreakOff()
  // Activate the ctrl-c handler;
  {
  static int isoff=0;
  if(isoff) return;
  SetConsoleCtrlHandler(&Break_Handler,1);
  isoff=1;
  }

int getkeyevent(int *shift, int *ctrl, int *vk)
  // Pick up keyboard events, called by getch() and can be used alone.
  // Return the ascii value of ascii keys and zero for other keys.
  // "shift" and "ctrl" are flags for upper case and the ctrl-key state,
  // "vk" ls the win32 VK_ scan code.
  {
  HANDLE h;
  INPUT_RECORD inrec;
  KEY_EVENT_RECORD r;
  DWORD s;
  int nr, k, shp, cpl, done;
  h=GetStdHandle(STD_INPUT_HANDLE);
  done=0;
  while(!done)
    {
    ReadConsoleInput(h,&inrec,1,&nr);
    if(inrec.EventType!=KEY_EVENT) continue;
    r=inrec.Event.KeyEvent;
    if(!r.bKeyDown) continue;
    k=r.wVirtualKeyCode;
    if( (k==VK_SHIFT) || (k==VK_CONTROL) || (k==VK_CAPITAL) ||
      (k==VK_NUMLOCK) || (k==VK_SCROLL) || (k==VK_SNAPSHOT) ||
      (k==VK_MENU) ) continue;
    s=r.dwControlKeyState;
    if((s & LEFT_ALT_PRESSED) || (s & RIGHT_ALT_PRESSED)) continue;
    shp=s & SHIFT_PRESSED;
    cpl=s & CAPSLOCK_ON;
    *shift=(shp && !cpl) || (!shp && cpl);
    *ctrl=(s & RIGHT_CTRL_PRESSED) || (s & LEFT_CTRL_PRESSED);
    *vk=k;
    done=1;
    }
  return r.uChar.AsciiChar;
  }

int EXP getch2(int *shift, int *ctrl, int *vk)
  // exported wrapper for getkeyevent()
  {
  return getkeyevent(shift, ctrl, vk);
  }

int EXP getch()
  // translate keyboard events
  // Return the ascii value if it is an ascii key.
  // Non-ascii keys require using getch() twice.  The first return is zero,
  // the second is the scan code as if from interrupt 16h function 10h.
  {
  static int Kbd_Queue=0;
  int ac, sh, ctr, vk, ret;
  if(Kbd_Queue) { ret=Kbd_Queue; Kbd_Queue=0; return ret; }
  ac=getch2(&sh,&ctr,&vk);
  if(sh&&(ac==9)) { Kbd_Queue=0x0F; return 0; }  // tab-left
  if(ac) return ac;  // ascii
  switch(vk)
    {
    case VK_LEFT: if(ctr) Kbd_Queue=0x73; else Kbd_Queue=0x4B; break;
    case VK_RIGHT: if(ctr) Kbd_Queue=0x74; else Kbd_Queue=0x4D; break;
    case VK_UP: Kbd_Queue=0x48; break;
    case VK_DOWN: Kbd_Queue=0x50; break;
    case VK_DELETE: Kbd_Queue=0x53; break;
    case VK_HOME: if(ctr) Kbd_Queue=0x77; else Kbd_Queue=0x47; break;
    case VK_END: if(ctr) Kbd_Queue=0x75; else Kbd_Queue=0x4F; break;
    case VK_PRIOR: Kbd_Queue=0x49; break;
    case VK_NEXT: Kbd_Queue=0x51; break;
    case VK_INSERT: Kbd_Queue=0x52; break;
    case VK_F1: Kbd_Queue=0x3B; break;
    case VK_F2: Kbd_Queue=0x3C; break;
    case VK_F3: Kbd_Queue=0x3D; break;
    case VK_F4: Kbd_Queue=0x3E; break;
    case VK_F5: Kbd_Queue=0x3F; break;
    case VK_F6: Kbd_Queue=0x40; break;
    case VK_F7: Kbd_Queue=0x41; break;
    case VK_F8: Kbd_Queue=0x42; break;
    case VK_F9: Kbd_Queue=0x43; break;
    case VK_F10: Kbd_Queue=0x44; break;
    case VK_F11: Kbd_Queue=0x85; break;
    case VK_F12: Kbd_Queue=0x86; break;
    default: Kbd_Queue='?';
    }
  return 0;
  }
